/**
* \file DisjointSet.cpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#include "DisjointSet.hpp"
#include "Debug.hpp"
#include <iostream>
#include <cstring>
#include <algorithm>
using namespace std;

DisjointSet::DisjointSet( unsigned int numItems ) :
	mNumItems( numItems )
{
	mParents = new unsigned int[mNumItems];
	mSizes = new unsigned int[mNumItems];

	// Each item is its own parent to begin with.
	// Since each set has one item (itself), its size should be 1.
	for( unsigned int i = 0; i < mNumItems; i++ )
	{
		mParents[i] = i;
		mSizes[i] = 1;
	}
}

DisjointSet::DisjointSet( const DisjointSet & that )
{
	// Allocate space to copy into.
	mParents = new unsigned int[that.mNumItems];
	mSizes = new unsigned int[that.mNumItems];

	// Copy the data from that into this.
	mNumItems = that.mNumItems;
	memcpy( mParents, that.mParents, sizeof( unsigned int ) * mNumItems );
	memcpy( mSizes, that.mSizes, sizeof( unsigned int ) * mNumItems );
}

DisjointSet::~DisjointSet()
{
	delete [] mParents;
	delete [] mSizes;
}

void DisjointSet::setUnion( unsigned int x, unsigned int y )
{
	// Can't union x with itself, must make sure that x and y won't cause
	// segfaults, and that x and y are actually in different sets.
	fatalAssert( x != y && x < mNumItems && y < mNumItems && find( x ) != find( y ), "Failure in setUnion()." );

	// If x isn't the representative of its set, make it so.
	if( !isRepresentative( x ) )
	{
		rebase( x );
	}
	// If y isn't the representative of its set, make it so.
	if( !isRepresentative( y ) )
	{
		rebase( y );
	}

	// If set x is bigger, make it the parent. This tries to loosely manage the
	// height of the structure so that rebase() and find() run in approximately
	// logarithmic time.
	if( mSizes[x] >= mSizes[y] )
	{
		mParents[y] = x;
		mSizes[x] += mSizes[y];
	}
	// Otherwise, set y was bigger, make it the parent.
	else
	{
		mParents[x] = y;
		mSizes[y] += mSizes[x];
	}
}

unsigned int DisjointSet::find( unsigned int x ) const
{
	// Avoid segfault.
	fatalAssert( x < mNumItems, "You tried to make find() cause a segfault." );

	// Follow the parents up until the representative is reached.
	unsigned int curr = x;
	while( !isRepresentative( curr ) )
	{
		curr = mParents[curr];
	}

	return curr;
}

void DisjointSet::split( unsigned int x, unsigned int y )
{
	// Avoid a segfault.
	fatalAssert( x != y && x < mNumItems && y < mNumItems, "Failure in split()." );
	// Can only split if the sets which containted x and y were unioned using
	// the call union( x, y ) or union( y, x ); this means that one is the
	// parent of the other, and we can use the connection between them as a
	// cut edge.
	fatalAssert( mParents[x] == y || mParents[y] == x, "Pointless to split two sets which are already disjoint" );

	// Determine which of x and y is the parent, and which is the child.
	unsigned int parent = ( mParents[x] == y ? y : x );
	unsigned int child = ( mParents[x] == y ? x : y );

	unsigned int curr = parent;
	while( !isRepresentative( curr ) )
	{
		// Must update the sizes all the way up.
		mSizes[curr] -= mSizes[child];
		curr = mParents[curr];
	}
	mSizes[curr] -= mSizes[x];

	// Make the child the representative of its set.
	mParents[child] = child;
}

void DisjointSet::getSetSizes( std::vector<unsigned int> & sizes ) const
{
	sizes.clear();
	for( unsigned int i = 0; i < mNumItems; i++ )
	{
		// If an item is a representative, add its size to the vector.
		if( isRepresentative( i ) )
		{
			sizes.push_back( mSizes[i] );
		}
	}

	// Sort it greatest to least.
	std::sort( sizes.begin(), sizes.end(), std::greater<int>() );
}

void DisjointSet::print() const
{
	cout << ">>> Set State <<< " << endl;
	for( unsigned int i = 0; i < mNumItems; i++ )
	{
		cout << "Item: " << i
			 << " (Parent: " << mParents[i]
			 << ", Size: " << mSizes[i]
			 << ", Representative: " << find( i ) << ")" << endl;
	}
}

DisjointSet & DisjointSet::operator=( const DisjointSet & that )
{
	if( this != &that )
	{
		// Free memory that was previously used.
		delete [] mParents;
		delete [] mSizes;

		// Allocate new space.
		mNumItems = that.mNumItems;
		mParents = new unsigned int[mNumItems];
		mSizes = new unsigned int[mNumItems];

		// Copy the data from that object.
		memcpy( mParents, that.mParents, sizeof( unsigned int ) * mNumItems );
		memcpy( mSizes, that.mSizes, sizeof( unsigned int ) * mNumItems );
	}

	return *this;
}

void DisjointSet::rebase( unsigned int x )
{
	// If x is already a representative, we should terminate because this is a
	// waste of a function call.
	fatalAssert( !isRepresentative( x ), "Pointless to rebase a representative." );

	// Essentially, this will boil down to a linked-list reversal, with some
	// additional size-adjusting logic at the end. Start with the reversal.
	unsigned int prev = x;
	unsigned int curr = x;
	unsigned int next = mParents[x];
	while( !isRepresentative( curr ) )
	{
		mParents[curr] = prev;
		prev = curr;
		curr = next;
		next = mParents[next];
	}
	mParents[curr] = prev;

	// Now, adjust the sizes of the elements
	unsigned int formerRepSize = mSizes[curr];
	while( !isRepresentative( curr ) )
	{
		next = mParents[curr];
		mSizes[curr] = formerRepSize - mSizes[next];
		curr = next;
	}
	mSizes[curr] = formerRepSize;
}
