/**
* \file DisjointSet.hpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#ifndef DISJOINT_SET_HPP
#define DISJOINT_SET_HPP

#include <vector>

/** \class DisjointSet
* \brief Represents a disjoint set of integers.
* \details Underneath the hood, it represents the disjoint set as a directed tree. This
* is to support computing the sizes of the connected components of a forest.
* We know that each edge of the tree is a cut edge, so adding an edge always
* unions two disjoint components, and removing an edge always splits one
* a component into two disjoint ones.
*/
class DisjointSet
{
	public:
		/** \brief Constructor.
		* Creates a disjoint set with numItems sets labelled 0 to numItems - 1.
		* Initially, each set is disjoint.
		* \param numItems Number of disjoint sets to start with.
		* \pre numItems should be greater than 0 to avoid weird memory errors.
		*/
		DisjointSet( unsigned int numItems );
		/** \brief Copy constructor.
		* Copies contents of that DisjointSet object into this object.
		* \param that DisjointSet object to copy.
		* \pre that should be a valid DisjointSet object.
		* \post This object will be an exact copy of that DisjointSet, and that
		* DisjointSet will not be modified at all.
		*/
		DisjointSet( const DisjointSet & that );
		/** \brief Destructor.
		* Frees the memory used by this object. Called automatically when the
		* object goes out of scope.
		*/
		~DisjointSet();

		/** \brief Unions the set which contains x with that which contains y.
		* \param x Item in a set. Doesn't need to be a representative.
		* \param y Item in a set. Doesn't need to be a representative.
		* \pre x and y should not be equal to each other, and they should both
		* satisfy 0 <= x,y < mNumItems. Also, x and y should not be in the same
		* set.
		* \post The two disjoint sets, one with x, and one with y, will become
		* a single set whose representative is x if the original disjoint set
		* containing x was larger, or y if the original disjoint set containing
		* y was larger.
		*/
		void setUnion( unsigned int x, unsigned int y );
		/** \brief Gets the representative (or label) of the set that contains
		* item x.
		* \param x Item in a set.
		* \pre Need 0 <= x < mNumItems.
		* \post Returns the representative of the set which contains x. Does
		* not modify this object at all.
		*/
		unsigned int find( unsigned int x ) const;
		/** \brief Splits the set which contains x and y into the disjoint sets
		* which originally contained x and y.
		* Essentially undoes the setUnion() operation.
		* \param x Item in a set. Doesn't need to be a representative.
		* \param y Item in a set. Doesn't need to be a representative.
		* \pre x and y should not be equal to each other, and they should both
		* satisfy 0 <= x,y < mNumItems. Also, x and y should be in the same set.
		* \post The set which contains x and y will be split into the two
		* disjoint sets which originally contained x and y, respectively.
		*/
		void split( unsigned int x, unsigned int y );

		/** \brief Makes a list of the cardinalites of each set.
		* This is sorted greatest to least and is identical to the sizes of the
		* connected components of the tree (the way this object is used, in the
		* main program, at least). The sizes of the connected components are
		* needed when computing the CSF. See Theorem 2.5 in "A Symmetric
		* Function Generalization of the Chromatic Polynomial of a Graph" by
		* Richard Stanley.
		* \param sizes Vector in which to store the set sizes.
		* \post Doesn't modify this object, but sizes will contain the
		* cardinalities of each set, sorted largest to smallest.
		*/
		void getSetSizes( std::vector<unsigned int> & sizes ) const;
		/** \brief Prints information useful for debugging.
		* For each item, it prints the items parent, representative, and the
		* size of the component if it was disconnected from its parent.
		* \post Doesn't modify this object at all.
		*/
		void print() const;

		/** \brief Overloaded assignment operator.
		* Enables deep copy of this object using assignment operator "=".
		* \param that The DisjointSet object to copy into this object.
		* \pre that should be a valid DisjointSet object.
		* \post This object will be an exact copy of that.
		*/
		DisjointSet & operator=( const DisjointSet & that );

	private:
		unsigned int mNumItems;	//< Total number of items in all sets.
		unsigned int *mParents;	//< Array such that item i's parent is stored at postion i.
		unsigned int *mSizes;	//< Array such position i contains the number of items that point to item i.

		/** \brief Determines whether x is a representative.
		* A representative is an item which is used as an identifier for its
		* set, and is marked by making the item its own parent.
		* \param x Item in a set.
		* \pre x should satisfy 0 <= x < mNumItems.
		* \post Doesn't modify this object at all.
		* \return True if x is a representative, false, otherwise.
		*/
		bool isRepresentative( unsigned int x ) const
		{
			return mParents[x] == x;
		};

		/** \brief Make x the representative of its set.
		* \param x Item in a set.
		* \pre x should satisfy 0 <= x < mNumItems, and x should not be a
		* representative.
		* \post The items in the set to which x belongs will be modified so
		* that x is the new representative of that set.
		*/
		void rebase( unsigned int x );
};

#endif

