/**
* \file SubsetDeltaGenerator.cpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#include "SubsetDeltaGenerator.hpp"
#include <cstddef>
#include <cstring>

SubsetDeltaGenerator::SubsetDeltaGenerator( unsigned int n )
{
	this->n = n;
	k = 0;
	// Don't allocate space, don't need it yet.
	mCurrSubset = NULL;
}

SubsetDeltaGenerator::~SubsetDeltaGenerator()
{
	if( mCurrSubset != NULL )
	{
		delete [] mCurrSubset;
	}
}

void SubsetDeltaGenerator::getDeltasForSubsetsOfFixedLength( unsigned int k, std::vector<SubsetDelta> & deltas )
{
	this->k = k;
	mDeltas.clear();

	// Free previously allocated space if there was any.
	if( mCurrSubset != NULL )
	{
		delete [] mCurrSubset;
		mCurrSubset = NULL;
	}

	// Allocate new space for the current subset. As described in the
	// header file, this array should be initalized to { 0, 1, 2, ..., k - 1 }.
	mCurrSubset = new unsigned int[k];
	for( unsigned int i = 0; i < k; i++ )
	{
		mCurrSubset[i] = i;
	}

	// Kick off the subset generation process, see the Eades/McKay paper
	// referenced in the header for more information.
	process( 1, 1 );
	forward( 1, 0 );

	// Set the vector passed by reference to this function to the deltas
	// just computed by the above 2 calls.
	deltas = mDeltas;
}

// See the Eades/McKay paper for more information.
void SubsetDeltaGenerator::forward( int pointer, int difference )
{
	if( ( pointer < (int)k ) && ( difference - pointer < (int)n - (int)k - 1 ) )
	{
		forward( pointer + 2, difference + 2 );
		process( pointer + 1, n - k + pointer + 1 );
		reverse( pointer + 1, difference + 2 );
		process( pointer, difference + 2 );
		forward( pointer, difference + 1 );
	}
	else if( pointer == (int)k )
	{
		for( int i = difference + 2; i <= (int)n; i++ )
		{
			process( k, i );
		}
	}
}

// See the Eades/McKay paper for more information.
void SubsetDeltaGenerator::reverse( int pointer, int difference )
{
	if( ( pointer < (int)k ) && ( difference - pointer < (int)n - (int)k - 1 ) )
	{
		reverse( pointer, difference + 1 );
		process( pointer, difference + 1 );
		forward( pointer + 1, difference + 2 );
		process( pointer + 1, difference + 2 );
		reverse( pointer + 2, difference + 2 );
	}
	else if( pointer == (int)k )
	{
		for( int i = (int)n - 1; i >= difference + 1; i-- )
		{
			process( k, i );
		}
	}
}

// See the Eades/McKay paper for more information.
void SubsetDeltaGenerator::process( int posChanged, int newValue )
{
	SubsetDelta d;
	d.oldValue = mCurrSubset[posChanged - 1];
	d.newValue = newValue - 1;
	mDeltas.push_back( d );

	mCurrSubset[posChanged - 1] = newValue - 1;
}

