/**
* \file SubsetDeltaGenerator.hpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#ifndef SUBSET_DELTA_GENERATOR_HPP
#define SUBSET_DELTA_GENERATOR_HPP

#include <vector>

/** \brief Describes how to change a subset to get the next subset.
* \details Given a subset of fixed length, if oldValue is replaced by newValue,
* this yields the next subset in the sequence.
*/
struct SubsetDelta
{
	unsigned int oldValue;
	unsigned int newValue;
};

/** \class SubsetDeltaGenerator.
* \brief Generates SubsetDelta structs for subsets of fixed length.
* \details Uses the algorithm described in the paper "An Algorithm for Generating
* Subsets of Fixed Size with a Strong Minimal Change Property" by Peter Eades
* and Brendan McKay.
*/
class SubsetDeltaGenerator
{
	public:
		/** \brief Constructor.
		* Initializes generator for subsets of fixed size of the set { 0, 1, 2,
		* 3, ..., n - 1 }.
		* \param n Number of items in the set to make subsets of.
		*/
		SubsetDeltaGenerator( unsigned int n );
		/** \brief Destructor.
		* Called automatically, cleans up memory used by the object.
		*/
		~SubsetDeltaGenerator();

		/** \brief Compute a list of minimum changes needed to change one
		* subset to the next in the list.
		* Starting with the subset  { 0, 1, 2, ..., k - 1 }, each SubsetDelta
		* struct indicates what item needs to be replaced, and what to replace
		* it with in order to get the next subset in the list.
		* \param k Desired fixed size for the subsets.
		* \param deltas Vector which will hold the deltas as described above.
		* \pre k should satisfy 0 <= k <= n.
		* \post deltas will contain the SubsetDeltas describing the minimum 
		* changes between subsets.
		*/
		void getDeltasForSubsetsOfFixedLength( unsigned int k, std::vector<SubsetDelta> & deltas );

	private:
		unsigned int		n;				//< Number of items in the set { 0, 1, ..., n - 1 }.
		unsigned int		k;				//< Size of the subsets.
		unsigned int		*mCurrSubset;	//< Array representing current subset under consideration.
		std::vector<SubsetDelta> mDeltas;	//< List of minimum changes needed to change one subset to the next.

		// These are ported directly from the Eades/McKay paper referenced
		// above. The only difference is that they use indexing relative to 0,
		// since this is more natural in C++.
		void forward( int pointer, int difference );
		void reverse( int pointer, int difference );
		void process( int posChanged, int newValue );
};

#endif
