/**
* \file Tree.cpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#include "Tree.hpp"
#include "Debug.hpp"
#include <vector>
#include <queue>
#include <algorithm>
#include <cstring>
#include <sstream>
#include <string>
#include <iostream>
using namespace std;

// Initializes an empty tree with no vertices.
Tree::Tree() :
	mOrder( 0 ),
	mDegrees( NULL ),
	mOffsets( NULL ),
	mAdjArray( NULL )
{
}

// Initializes a tree of the given order (number of vertices), using the level
// sequence array.
Tree::Tree( unsigned int order, int *levelSequence ) :
	mOrder( order )
{
	// Allocate and initialize member arrays.
	mOffsets = new unsigned int[mOrder];
	mDegrees = new unsigned int[mOrder];
	memset( mOffsets, 0, sizeof( unsigned int ) * mOrder );
	memset( mDegrees, 0, sizeof( unsigned int ) * mOrder );

	// Determine the edges of the tree.
	vector<Edge> edges;
	Edge e;
	for( unsigned int i = 2; i <= mOrder; i++ )
	{
		e.u = i - 1;
		e.v = levelSequence[i - 1] - 1;
		mDegrees[e.u]++;
		mDegrees[e.v]++;
		edges.push_back( e );
	}

	// Initialize the offsets array.
	for( unsigned int i = 1; i < mOrder; i++ )
	{
		mOffsets[i] = mOffsets[i - 1] + mDegrees[i - 1];
	}

	// Initialize the adjacency array.
	mAdjArray = new unsigned int[2 * ( mOrder - 1 )];
	unsigned int *currSizes = new unsigned int[mOrder];
	memset( currSizes, 0, sizeof( unsigned int ) * mOrder );
	for( unsigned int i = 0; i < edges.size(); i++ )
	{
		unsigned int u = edges[i].u;
		unsigned int v = edges[i].v;
		mAdjArray[mOffsets[u] + currSizes[u]++] = v;
		mAdjArray[mOffsets[v] + currSizes[v]++] = u;
	}
	delete [] currSizes;
}

Tree::Tree( const Tree & that )
{
	// Copy the order by value, and allocate space for the arrays.
	mOrder = that.mOrder;
	mDegrees = new unsigned int[mOrder];
	mOffsets = new unsigned int[mOrder];
	mAdjArray = new unsigned int[2 * ( mOrder - 1 )];

	// Copy the arrays from that into this object.
	memcpy( mDegrees, that.mDegrees, sizeof( unsigned int ) * mOrder );
	memcpy( mOffsets, that.mOffsets, sizeof( unsigned int ) * mOrder );
	memcpy( mAdjArray, that.mAdjArray, sizeof( unsigned int ) * ( 2 * ( mOrder - 1 ) ) );
}

Tree::~Tree()
{
	// Clean up the arrays if they have been allocated space.
	if( mDegrees != NULL )
	{
		delete [] mDegrees;
	}
	if( mOffsets != NULL )
	{
		delete [] mOffsets;
	}
	if( mAdjArray != NULL )
	{
		delete [] mAdjArray;
	}
}

std::vector<Edge> Tree::getEdges() const
{
	std::vector<Edge> edges;
	for( unsigned int i = 0; i < mOrder; i++ )
	{
		for( unsigned int j = 0; j < mDegrees[i]; j++ )
		{
			Edge e;
			e.u = i;
			e.v = mAdjArray[mOffsets[i] + j];
			// Don't want duplicate edges in the list
			if( e.u < e.v )
			{
				edges.push_back( e );
			}
		}
	}

	return edges;
}

void Tree::getDegreeSequence( std::vector<unsigned int> & degreeSequence ) const
{
	// Just duplicate the mDegrees array and sort it.
	degreeSequence.clear();
	degreeSequence.insert( degreeSequence.end(), &mDegrees[0], &mDegrees[mOrder] );
	std::sort( degreeSequence.begin(), degreeSequence.end(), std::greater<int>() );
}

void Tree::getPathNums( std::vector<unsigned int> & pathNums ) const
{
	// Initialize pathNums vector to all 0's.
	pathNums.clear();
	for( unsigned int i = 0; i < mOrder; i++ )
	{
		pathNums.push_back( 0 );
	}

	// Allocate visited array.
	bool *visited = new bool[mOrder];
	for( unsigned int i = 0; i < mOrder; i++ )
	{
		// Reset array so that no vertex has been visited.
		memset( visited, 0, sizeof( bool ) * mOrder );
		// Start DFS on current vertex (for all vertices).
		pathNumsDfs( i, 1, visited, pathNums );
	}
	delete [] visited;

	// Divide each count by two since each path is counted twice.
	for( unsigned int i = 0; i < pathNums.size(); i++ )
	{
		pathNums[i] /= 2;
	}
}

void Tree::print() const
{
	vector<Edge> edgeSet = getEdges();
	for( unsigned int i = 0; i < edgeSet.size(); i++ )
	{
		cout << "(" << edgeSet[i].u << ", " << edgeSet[i].v << ") ";
	}
	cout << endl;
}

Tree & Tree::operator=( const Tree & that )
{
	// Prevent self-assignment, it causes problems. Otherwise, free, reallocate, and copy.
	if( this != &that )
	{
		delete [] mDegrees;
		delete [] mOffsets;
		delete [] mAdjArray;

		mOrder = that.mOrder;

		mDegrees = new unsigned int[mOrder];
		mOffsets = new unsigned int[mOrder];
		mAdjArray = new unsigned int[2 * ( mOrder - 1 )];

		memcpy( mDegrees, that.mDegrees, sizeof( unsigned int ) * mOrder );
		memcpy( mOffsets, that.mOffsets, sizeof( unsigned int ) * mOrder );
		memcpy( mAdjArray, that.mAdjArray, sizeof( unsigned int ) * ( 2 * ( mOrder - 1 ) ) );
	}

	return *this;
}

void Tree::pathNumsDfs( unsigned int i, unsigned int count, bool *& visited, std::vector<unsigned int> & pathNums ) const
{
	// Mark current vertex as visited.
	visited[i] = true;
	for( unsigned int j = 0; j < mDegrees[i]; j++ )
	{
		unsigned int adjVert = mAdjArray[mOffsets[i] + j];
		// If the next adjacent vertex hasn't been visited already.
		if( !visited[adjVert] )
		{
			// Increment the number of paths of this length.
			pathNums[count]++;
			// Recurse on the adjacent vertex, and make the path length one unit longer.
			pathNumsDfs( adjVert, count + 1, visited, pathNums );
		}
	}
}

