/**
* \file Tree.hpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#ifndef TREE_HPP
#define TREE_HPP

#include <vector>
#include <string>

/** \brief An undirected edge between vertices u and v.
*/
struct Edge
{
	unsigned int u;
	unsigned int v;
};

/** \class Tree
* \brief A labelled, free tree.
* \details The labels are arbitrary, but useful for
* operating on the tree.
*/
class Tree
{
	public:
		/** \brief Default constructor.
		* Creates an empty Tree, one with no vertices or edges.
		*/
		Tree();

		/** \brief Constructor.
		* Creates a Tree of the given order (i.e. number of vertices), and the
		* given level sequence (see TreeGenerator.hpp and TreeGenerator.cpp for
		* details). Essentially, the level sequence is an array which describes
		* which edges the Tree contains.
		* \param order The number of vertices in the Tree. Should also be the
		* size of the levelSequence array.
		* \param levelSequence Array which describes the structure of the Tree.
		* See the TreeGenerator class for more information on how this array is
		* defined and created.
		* \pre The order should be equal to the size of the levelSequence array.
		* \post The level sequence is unchanged, and this object is a Tree with
		* edges as described by the level sequence array.
		*/
		Tree( unsigned int order, int *levelSequence );

		/** \brief Copy constructor.
		* Copies that Tree into this object.
		* \param that The Tree to copy into this object.
		* \pre that should be a valid Tree object.
		* \post This object will be an exact copy in memory of that Tree, and
		* that object will not be modified at all.
		*/
		Tree( const Tree & that );

		/** \brief Destructor.
		* Automatically called when this Tree goes out of scope. Frees the
		* memory this Tree takes up.
		*/
		~Tree();

		/** \brief Get the number of vertices in this Tree.
		* \post Doesn't modify this Tree object at all.
		* \return Positive integer number of vertices in this Tree.
		*/
		unsigned int getOrder() const { return mOrder; };
		/** \brief Get the degree of the given vertex v.
		* Degree is the number of edges incident on that vertex.
		* \param v The vertex you want to know the degree of.
		* \pre v should be a positive integer, 0 <= v < n, where n is the order
		* of this Tree. This is because the vertex labels range from 0 to n - 1.
		* \post Doesn't modify this Tree object at all.
		* \return Positive integer number of edges incident with vertex v.
		* Returns 0 if the precondition for v is not met.
		*/
		unsigned int getDegree( unsigned int v ) const { if( v < mOrder ) return mDegrees[v]; return 0; };

		/** \brief Get a list of edges in this Tree.
		* Gets a vector containing each undirected edge present in this Tree,
		* stored as Edge structs.
		* \post Doesn't modify this Tree object at all.
		* \return A vector (i.e. an STL dynamic array) which contains the edges
		* of this Tree in no particular order.
		*/
		std::vector<Edge> getEdges() const;
		/** \brief Gets the degree sequence of this Tree.
		* Sorts the degree sequence (a list of positive integers) from largest
		* to smallest.
		* \param degreeSequence This vector will be cleared and its new
		* contents will be the sorted degree sequence.
		* \post Doesn't modify this Tree object at all.
		*/
		void getDegreeSequence( std::vector<unsigned int> & degreeSequence ) const;
		/** \brief Gets the path sequence of this Tree.
		* List of positive integers, each of which is the number of paths of
		* the length given by its position in the vector (first position is
		* length 0, second position is length 1, etc.). This function is a
		* wrapper for the private function pathNumsDfs().
		* \param pathNums This vector will be cleared and its new contents will
		* be the path sequence as described above.
		* \post Doesn't modify this Tree object at all.
		*/
		void getPathNums( std::vector<unsigned int> & pathNums ) const;

		/** \brief Prints a list of the edges of this Tree.
		* \post Doesn't modify this Tree object at all.
		*/
		void print() const;

		/** \brief Overloaded assignment operator.
		* Manages memory when assignment operator is used with this oject.
		* E.g. tree1 = tree2;
		* \param rhs Tree on the right hand side of the equals sign.
		* \pre rhs should be a valid Tree object.
		* \post This Tree will be an exact duplicate of the Tree rhs.
		*/
		Tree & operator=( const Tree & rhs );
	
	private:
		/** \brief Counts the number of paths of a given length in the tree,
		* for all possible path lengths in the tree. Uses a depth first search,
		* hence the name DFS. See getPathNums() for how to use it.
		* \param i Current vertex being considered by the depth first search.
		* \param count Length of the current path being considered.
		* \param visited Array of booleans, marking each vertex as visited or
		* not: true == visted, false == unvisted.
		* \param pathNums This vector will contain the path sequence given a
		* single starting point. To get the entire path sequence, this function
		* must be called once for each starting vertex, as in getPathNums().
		* \pre i should be a vertex in the tree; visited should be an array of
		* all false values (for the first call, i.e. before it recurses) of
		* length mOrder.
		* \post Doesn't modify this Tree object at all.
		*/
		void pathNumsDfs( unsigned int i, unsigned int count, bool *& visited, std::vector<unsigned int> & pathNums ) const;

		unsigned int	mOrder;		//< Number of vertices in the tree.
		unsigned int	*mDegrees;	//< Degrees of each vertex, from vertex 0 to mOrder - 1.
		unsigned int	*mOffsets;	//< Offsets to the location of the adjacency lists in mAdjArray.
		unsigned int	*mAdjArray;	//< Array containing adjacency lists.
};

#endif

