/**
* \file TreeGenerator.hpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#ifndef TREE_GENERATOR_HPP
#define TREE_GENERATOR_HPP

#include "Tree.hpp"

/** \class TreeGenerator
* \brief Generates non-isomorphic trees of a given order.
* \details Ported from the Sage source for the TreeIterator object which was
* written by Ryan Dingman. His code was an implementation of the algorithm
* described in the paper "Constant Time Generation of Free Trees" by Robert
* Alan Wright, Bruce Richmond, Andrew Odlyzko, and Brendan McKay.
*/
class TreeGenerator
{
	public:
		/** \brief Constructor.
		* Initializes the object to generate trees of the given order.
		* \param numVertices Desired order, or number of vertices, for the
		* trees to generate.
		*/
		TreeGenerator( unsigned int numVertices );
		/** \brief Destructor.
		* Called automatically, cleans up the memory used by this object.
		*/
		~TreeGenerator();

		// Get the next tree generated, returns true until there are no more trees
		// with the specified number of vertices. If false, the tree t won't necessarily
		// contain anything meaningful, so take care using it in such a case.
		/** \brief Get the next tree generated by the algorithm.
		* \param t Passed by reference to store the tree created by the
		* function call.
		* \post t will be a tree non-isomorphic to any other tree generated by
		* this object if the function returns true. Otherwise, t will contain
		* invalid data.
		* \return True if the tree is valid, false if there were no more trees
		* left to generate.
		*/
		bool nextTree( Tree & t );

	private:
		unsigned int	mNumVertices;
		int				*L;
		int				*currentLevelSequence;

		int				p;
		int				q;
		int				h1;
		int				h2;
		int				c;
		int				r;

		bool			firstTime;

		void generateFirstLevelSequence();
		void generateNextLevelSequence();
};

#endif
