/**
* \file main.cpp
* \author Keeler M. Russell, keeler.russell@gmail.com
* \note You are free to use or modify this code to your heart's desire.
*/

#include <pthread.h>
#include <stdio.h>
#include <vector>
#include <queue>
#include <map>
#include <set>
#include <algorithm>
#include <utility>
#include <sstream>
#include <cmath>
#include "Tree.hpp"
#include "TreeGenerator.hpp"
#include "DisjointSet.hpp"
#include "SubsetDeltaGenerator.hpp"
#include "Debug.hpp"
#include <iostream>
using namespace std;

typedef map< vector<unsigned int>, vector<Tree> > TreeByDegSeqHash;
typedef map< pair< vector<unsigned int>, vector<unsigned int> >, vector<Tree> > DegSeqPathNumHash;
typedef map< vector<unsigned int>, int> kSlice;
typedef vector< vector<SubsetDelta> > SubsetDeltaTable;
typedef queue< vector<Tree> > TreeQueue;

// Used to group trees according to degree and path sequences. Maps a degree
// sequence, path sequence pair to a group of trees with that degree
// degree sequence and path sequence.
DegSeqPathNumHash	treeHash;
// Queue of groups of trees for the worker threads to pull from. Each group has
// trees with the same degree and path sequences, but the degree and path
// sequences of each group are different from those of other groups. Also, each
// group must have at least two trees, since if a group has a single tree, that
// tree can't have the same CSF as any other tree of the given order, as the 
// paper "On Distinguishing Trees by their Chromatic Symmetric Functions" by
// Martin, Morin, and Wagner shows.
TreeQueue			treeQueue;
// Holds lists of deltas for subsets of fixed size. The list at index i holds
// the deltas for subsets of size i.
SubsetDeltaTable	subsetDeltaTable;
// Order of the trees, i.e. the number of vertices.
unsigned int		n;
// Number of threads to use. More is generally better.
unsigned int		numThreads;
// Pointer to array of thread handles.
pthread_t			*myThreads;
// Mutex for accessing the treeQueue data structure. Used to prevent concurrent
// access/modification of treeQueue by worker() threads.
pthread_mutex_t		treeQueueMutex;
// Mutex for using standard output in worker() threads.
pthread_mutex_t		outputMutex;

/** \brief Fills treeHash with trees. It maps a degree sequence, path sequence
* pair to a group of trees which share that degree sequence and path sequence.
* \param n Order (i.e. number of vertices) of trees to create.
*/
void getTrees( unsigned int n );
/** \brief Function that each thread executes.
* \details Computes the k-slice (or partial CSF) of each group of trees in
* the treeHash data structure, differentiating them based on CSF coefficients.
* \param arg Argmuents to the function cast as a void pointer. In this case,
* the thread number.
*/
void *worker( void *arg );
/** \brief Workhorse of the program.
* \details Generates the trees, computes the subsets needed to compute the CSF,
* and kicks of the CSF computation threads.
* \param n Order (number of vertices) of the trees to create and test.
*/
void testConjecture( unsigned int n );

int main( int argc, char **argv )
{
	// Correct user if they use the program incorrectly.
	if( argc != 3 )
	{
		cout << "Usage: " << argv[0] << " <number of vertices> <number of cores to use>" << endl;
		cout << "    Use the command nproc to find how many cores you have available." << endl;
		return 0;
	}

	// Determine the order of trees and number of threads to use.
	n = atoi( argv[1] );
	numThreads = atoi( argv[2] );
	myThreads = new pthread_t[numThreads];

	// Run the workhorse.
	testConjecture( n );

	// Clean up and exit.
	delete [] myThreads;
	pthread_exit( NULL );
	return 0;
}

void getTrees( unsigned int n )
{
	Tree t;
	TreeGenerator treeGen( n );
	vector<unsigned int> degSeq;
	vector<unsigned int> pathNums;

	// While there are still trees left to generate...
	while( treeGen.nextTree( t ) )
	{
		// Get the degree and path sequences of the current tree, and then add
		// the current tree to the corresponding vector in the treeHash table
		// based on those degree and path sequences.
		t.getDegreeSequence( degSeq );
		t.getPathNums( pathNums );
		treeHash[ make_pair( degSeq, pathNums ) ].push_back( t );
	}
}

// This is what all of the threads execute.
void *worker( void *arg )
{
	// Run the thread until this flag is set to false.
	bool threadRunning = true;
	// Current group of trees to work with.
	vector<Tree> treeList;
	while( threadRunning )
	{
		// Prevent concurrent access/modification of the tree queue to avoid
		// synchronization issues. This is the critical section.
		pthread_mutex_lock( &treeQueueMutex );
			// If the tree queue isn't empty, get the next group of trees.
			if( !treeQueue.empty() )
			{
				treeList = treeQueue.front();
				treeQueue.pop();
			}
			// Otherwise, this thread should be done running.
			else
			{
				pthread_mutex_unlock( &treeQueueMutex );
				break;
			}
		pthread_mutex_unlock( &treeQueueMutex );

		// The value of k is the size of the edge subsets. Can't use 0, 1, or
		// 2 because they don't differentiate the CSF of any trees with the
		// same order.
		unsigned int k = 3;
		while( k < n and treeList.size() > 1 )
		{
			vector<kSlice> kSliceList;

			// The power sum symmetric function form of the CSF equation
			// uses -1 to the cardinality of S as the sign for the summand. In
			// this function, we already know the cardinality of S, since it's
			// fixed as k.
			int additive = ( k % 2 == 0 ) ? 1 : -1;

			vector<SubsetDelta> deltas;
			deltas = subsetDeltaTable[k];	// Subset deltas for the given subset size.
	
			for( vector<Tree>::const_iterator it = treeList.begin();
				 it != treeList.end();
				 ++it )
			{
				vector<Edge> edgeSet = it->getEdges();
				kSlice currSlice;

				// We will use a disjoint set to track the sizes of the
				// connected components.
				DisjointSet components( n );

				// Initialize the connected components. We use the items in the
				// subsets given by the subset generator as the indices to
				// access the edge set vector. Since the subset generator
				// starts with the set { 0, 1, 2, ..., k - 1 }, we use these
				// indices for the edges in the first component to consider.
				for( unsigned int i = 0; i < k; i++ )
				{
					components.setUnion( edgeSet[i].u, edgeSet[i].v );
				}

				// lambdaOfS is the partition formed by sorting the sizes of
				// the connected components of the tree. See Thm. 2.5 in the
				// paper by Stanley referenced in Tree.hpp.
				vector<unsigned int> lambdaOfS;
				components.getSetSizes( lambdaOfS );
				currSlice[lambdaOfS] += additive;

				// Now, for each subset delta, remove the edge at the specified
				// index, and add the edge at the specified index. This is the
				// same as unioning the disjoint sets and splitting the
				// disjoint sets, respectively because each edge in a tree is a
				// cut edge. Then, compute lambdaOfS as above.
				for( unsigned int i = 1; i < deltas.size(); i++ )
				{
					Edge removeEdge = edgeSet[deltas[i].oldValue];
					Edge addEdge = edgeSet[deltas[i].newValue];
	
					components.split( removeEdge.u, removeEdge.v );
					components.setUnion( addEdge.u, addEdge.v );
	
					components.getSetSizes( lambdaOfS );
					currSlice[lambdaOfS] += additive;
				}

				// Save this k-slice.
				kSliceList.push_back( currSlice );
			}

			// Now, try to differentiate trees based on the coefficients of the
			// k-slices computed in the above loop. If even one coefficient
			// differs, the CSFs must be different.
			set<unsigned int> treesToKeep;
			// Nested for loop means O( n^2 ) performance in this loop, but the
			// indices i and j are initialized so that each k-slice is compared
			// to the others in the list only once.
			for( unsigned int i = 0; i < kSliceList.size(); i++ )
			{
				for( unsigned int j = i + 1; j < kSliceList.size(); j++ )
				{
					kSlice slice1 = kSliceList[i];
					kSlice slice2 = kSliceList[j];
					bool equal = true;
					kSlice::iterator it1 = slice1.begin();
					kSlice::iterator it2 = slice2.begin();
					// Iterate through the coefficients of the two k-slices and
					// if there is even one coefficient that doesn't match, the
					// k-slices aren't equal, so mark equal as false.
					while( it1 != slice1.end() && it2 != slice2.end() )
					{
						if( it1->first != it2->first ||
							it1->second != it2->second )
						{
							equal = false;
							break;
						}
						++it1;
						++it2;
					}
					// If the k-slices were equal, we'll keep the trees which
					// match up to these k-slices. Otherwise, discard them.
					if( equal )
					{
						treesToKeep.insert( i );
						treesToKeep.insert( j );
					}
				}
			}

			// Map the indices we saved above to actual tree objects.
			vector<Tree> treesKept;
			for( set<unsigned int>::iterator it = treesToKeep.begin();
				 it != treesToKeep.end();
				 ++it )
			{
				treesKept.push_back( treeList[*it] );
			}

			// Set the tree list to the trees kept after discarding the ones
			// with differing coefficients. Increment k to go to the next batch
			// of subsets to try to differentiate the trees based on the
			// coefficients from those.
			treeList = treesKept;
			k++;
		}

		// If there are still multiple trees with the same k-slices, then we've
		// tried every k-slice and at least two trees had all the same
		// coefficients. In other words, a counter example to Stanley's
		// conjecture has been found, so we print out the trees in question.
		if( treeList.size() > 1 )
		{
			// Only want one thread outputting at a time.
			pthread_mutex_lock( &outputMutex );
				cout << "COUNTEREXAMPLE FOUND!" << endl;
				for( unsigned int i = 0; i < treeList.size(); i++ )
				{
					treeList[i].print(); 
				}
			pthread_mutex_unlock( &outputMutex );
		}
	}

	pthread_exit( NULL );
}

void testConjecture( unsigned int n )
{
	getTrees( n );	// Generate the trees

	// 	Create the subset table.
	SubsetDeltaGenerator subsetGen( n - 1 );
	vector<SubsetDelta> temp;
	// First three entries are dummy vectors. We will not use them in worker()
	// anyway, since k-slices using k = 0, 1, or 2 don't differentiate trees
	// of the same order.
	for( unsigned int i = 0; i < 3; i++ )
	{
		subsetDeltaTable.push_back( temp );
	}
	// For larger size subsets, compute the deltas for subsets of size k, and
	// store that list at index k, for 3 <= k <= n.
	for( unsigned int i = 3; i < n; i++ )
	{
		subsetGen.getDeltasForSubsetsOfFixedLength( i, temp );
		subsetDeltaTable.push_back( temp );
	}

	// Build the tree queue from the tree hash, removing cohorts of trees of
	// size 1 from consideration by not adding them to the tree queue. The tree
	// queue will be used by worker() as a pool from which to grab groups of
	// trees to work on.
	for( DegSeqPathNumHash::const_iterator it = treeHash.begin();
		 it != treeHash.end();
		 ++it )
	{
		if( it->second.size() > 1 )
		{
			treeQueue.push( it->second );
		}
	}

	// Empty the tree hash, we don't need it anymore.
	treeHash.clear();

	// Initialize thread attributes and mutices.
	pthread_attr_t	threadAttributes;
	pthread_mutex_init( &treeQueueMutex, NULL );
	pthread_mutex_init( &outputMutex, NULL );
	pthread_attr_init( &threadAttributes );
	pthread_attr_setdetachstate( &threadAttributes, PTHREAD_CREATE_JOINABLE );

	// Create the worker threads.
	for( unsigned long i = 0; i < numThreads; i++ )
	{
		pthread_create( &myThreads[i], &threadAttributes, worker, (void *)i );
	}

	// Wait for the threads to finish.
	void *status;
	for( unsigned int i = 0; i < numThreads; i++ )
	{
		pthread_join( myThreads[i], &status );
	}

	// Clean up.
	pthread_attr_destroy( &threadAttributes );
	pthread_mutex_destroy( &treeQueueMutex );
	pthread_exit( NULL );
}
